﻿courseplay_manager = {};
local courseplay_manager_mt = Class(courseplay_manager);

function courseplay_manager:loadMap(name)
	if courseplay.isDeveloper then
		addConsoleCommand('cpAddMoney', ('Add %s to your bank account'):format(g_i18n:formatMoney(5000000)), 'devAddMoney', self);
		addConsoleCommand('cpAddFillLevels', 'Add 500\'000 l to all of your silos', 'devAddFillLevels', self);
	end;

	if g_currentMission.cp_courses == nil then
		--courseplay:debug("cp courses was nil and initialized", 8);
		g_currentMission.cp_courses = {};
		g_currentMission.cp_folders = {};
		g_currentMission.cp_sorted = {item={}, info={}};

		if g_server ~= nil and next(g_currentMission.cp_courses) == nil then
			courseplay_manager:load_courses()
			-- courseplay:debug(tableShow(g_currentMission.cp_courses, "g_cM cp_courses", 8), 8);
		end
	end;

	self.buttons = {};

	-- GLOBAL INFO TEXT
	self.globalInfoTextMaxNum = 20;
	self.globalInfoTextOverlays = {};
	self.buttons.globalInfoText = {};
	local git = courseplay.globalInfoText;
	for i=1,self.globalInfoTextMaxNum do
		local posY = git.backgroundPosY + (i - 1) * git.lineHeight;
		self.globalInfoTextOverlays[i] = Overlay:new(string.format("globalInfoTextOverlay%d", i), git.backgroundImg, git.backgroundPosX, posY, 0.1, git.buttonHeight);
		self:createButton('globalInfoText', 'goToVehicle', i, 'pageNav_7.png', git.buttonPosX, posY, git.buttonWidth, git.buttonHeight);
	end;
	self.buttons.globalInfoTextClickArea = {
		x1 = git.buttonPosX;
		x2 = git.buttonPosX + git.buttonWidth;
		y1 = git.backgroundPosY,
		y2 = git.backgroundPosY + (self.globalInfoTextMaxNum  * git.lineHeight);
	};

	self:createInitialCourseplayFile();

	self.playerOnFootMouseEnabled = false;
	self.wasPlayerFrozen = false;

	-- FIELD SCAN INFO DISPLAY
	self.fieldScanInfo = {};
	self.fieldScanInfo.fileW = 512/1080 / g_screenAspectRatio; --512/1920;
	self.fieldScanInfo.fileH = 256/1080;
	self.fieldScanInfo.contentW = 426/1080 / g_screenAspectRatio; --426/1920;
	self.fieldScanInfo.contentH = 180/1080;
	self.fieldScanInfo.bgX, self.fieldScanInfo.bgY = 0.5 - self.fieldScanInfo.contentW/2, 0.5 - self.fieldScanInfo.contentH/2;
	local bgPath = Utils.getFilename('img/fieldScanInfoBackground.png', courseplay.path);
	self.fieldScanInfo.bgOverlay = Overlay:new('fieldScanInfoBackground', bgPath, self.fieldScanInfo.bgX, self.fieldScanInfo.bgY, self.fieldScanInfo.fileW, self.fieldScanInfo.fileH);
	local xPadding = 20/1080 / g_screenAspectRatio; --20/1920;
	self.fieldScanInfo.lineX  = self.fieldScanInfo.bgX + xPadding;
	self.fieldScanInfo.line1Y = self.fieldScanInfo.bgY + 67/1080;
	self.fieldScanInfo.line2Y = self.fieldScanInfo.bgY + 40/1080;
	self.fieldScanInfo.loadX  = self.fieldScanInfo.bgX + 340/1080 / g_screenAspectRatio; --340/1920;
	self.fieldScanInfo.loadY  = self.fieldScanInfo.bgY + 57/1080;

	local loadingPath = Utils.getFilename('img/fieldScanInfoLoading.png', courseplay.path);
	self.fieldScanInfo.loadOverlay = Overlay:new('fieldScanInfoLoad', loadingPath, self.fieldScanInfo.loadX, self.fieldScanInfo.loadY, 32/1080 / g_screenAspectRatio, 32/1080);
	self.fieldScanInfo.loadRotStep = 0;
	self.fieldScanInfo.loadRotAdd = math.rad(-360/72); --rotate 5° to the right each step
	self.fieldScanInfo.rotationTime = 1000/72; --in ms

	self.fieldScanInfo.progressBarWidth = self.fieldScanInfo.contentW - 2 * xPadding - 5/1920;
	local progressBarBgPath = Utils.getFilename('img/progressBarBackground.png', courseplay.path);
	self.fieldScanInfo.progressBarBgOverlay = Overlay:new('fieldScanInfoProgressBarBg', progressBarBgPath, self.fieldScanInfo.lineX, self.fieldScanInfo.bgY + 16/1080, self.fieldScanInfo.progressBarWidth, 16/1080);
	local progressBarPath = Utils.getFilename('img/progressBar.png', courseplay.path);
	self.fieldScanInfo.progressBarOverlay = Overlay:new('fieldScanInfoProgressBar', progressBarPath, self.fieldScanInfo.lineX, self.fieldScanInfo.bgY + 16/1080, self.fieldScanInfo.progressBarWidth, 16/1080);
	self.fieldScanInfo.percentColors = {
		{ pct = 0.0, color = { 0.882353, 0.105882, 0 } },
		{ pct = 0.5, color = { 1.000000, 0.800000, 0 } },
		{ pct = 1.0, color = { 0.537255, 0.952941, 0 } }
	};


	if g_server ~= nil then
		courseplay.fields:loadAllCustomFields();
	end;

	courseplay.wageDifficultyMultiplier = Utils.lerp(0.5, 1, (g_currentMission.missionStats.difficulty - 1) / 2);

	self.firstRun = true;

	g_currentMission.environment:addMinuteChangeListener(courseplay_manager);
end;

function courseplay_manager:getSavegameFolder(index)
	if index then
		-- print(('getSavegameFolder(%d): "%ssavegame%d/"'):format(tostring(index), tostring(getUserProfileAppPath()), index));
		return ('%ssavegame%d/'):format(tostring(getUserProfileAppPath()), index);
	end;

	if g_currentMission.missionInfo.savegameDirectory then
		-- print(('getSavegameFolder(): g_currentMission.missionInfo.savegameDirectory=%q'):format(tostring(g_currentMission.missionInfo.savegameDirectory)));
		return g_currentMission.missionInfo.savegameDirectory;
	end;

	if g_currentMission.missionInfo.savegameIndex then
		-- print(('getSavegameFolder(): "%ssavegame%d/"'):format(tostring(getUserProfileAppPath()), g_currentMission.missionInfo.savegameIndex));
		return ('%ssavegame%d/'):format(tostring(getUserProfileAppPath()), g_currentMission.missionInfo.savegameIndex);
	end;

	return;
end;

function courseplay_manager:createInitialCourseplayFile()
	local filePath = courseplay.cpXmlFilePath;
	if filePath then
		local file;
		if not fileExists(filePath) then
			file = createXMLFile('courseplayFile', filePath, 'XML');
			-- print(string.format('\tcreateXmlFile("courseplayFile", [path], XML), file=%s', tostring(file)));
		else
			file = loadXMLFile('courseplayFile', filePath);
			-- print(string.format('\tloadXMLFile("courseplayFile", [path]), file=%s', tostring(file)));
		end;

		-- NOTE: usually "hasXMLProperty" would be used in this case. This presented some weird shitty problem though, in that the first of each set (posX, posX and debugScannedFields) were created, but not the second ones. So, getXML... is the better choice in this case.
		if not getXMLFloat(file, 'XML.courseplayHud#posX') then
			setXMLString(file, 'XML.courseplayHud#posX', ('%.3f'):format(courseplay.hud.infoBasePosX));
		end;
		if not getXMLFloat(file, 'XML.courseplayHud#posY') then
			setXMLString(file, 'XML.courseplayHud#posY', ('%.3f'):format(courseplay.hud.infoBasePosY));
		end;

		--[[ NO MORE CUSTOM POSITIONS FOR GLOBALINFOTEXT
		if not getXMLFloat(file, 'XML.courseplayGlobalInfoText#posX') then
			setXMLString(file, 'XML.courseplayGlobalInfoText#posX', ('%.3f'):format(courseplay.globalInfoText.posX));
		end;
		if not getXMLFloat(file, 'XML.courseplayGlobalInfoText#posY') then
			setXMLString(file, 'XML.courseplayGlobalInfoText#posY', ('%.3f'):format(courseplay.globalInfoText.posY));
		end;
		]]

		if getXMLBool(file, 'XML.courseplayFields#automaticScan') == nil then
			setXMLString(file, 'XML.courseplayFields#automaticScan', tostring(courseplay.fields.automaticScan));
		end;
		if getXMLBool(file, 'XML.courseplayFields#onlyScanOwnedFields') == nil then
			setXMLString(file, 'XML.courseplayFields#onlyScanOwnedFields', tostring(courseplay.fields.onlyScanOwnedFields));
		end;
		if getXMLBool(file, 'XML.courseplayFields#debugScannedFields') == nil then
			setXMLString(file, 'XML.courseplayFields#debugScannedFields', tostring(courseplay.fields.debugScannedFields));
		end;
		if getXMLBool(file, 'XML.courseplayFields#debugCustomLoadedFields') == nil then
			setXMLString(file, 'XML.courseplayFields#debugCustomLoadedFields', tostring(courseplay.fields.debugCustomLoadedFields));
		end;
		if not getXMLInt(file, 'XML.courseplayFields#scanStep') then
			setXMLInt(file, 'XML.courseplayFields#scanStep', courseplay.fields.scanStep);
		end;

		if getXMLBool(file, 'XML.courseplayWages#active') == nil then
			setXMLString(file, 'XML.courseplayWages#active', tostring(courseplay.wagesActive));
		end;
		if not getXMLInt(file, 'XML.courseplayWages#wagePerHour') then
			setXMLInt(file, 'XML.courseplayWages#wagePerHour', courseplay.wagePerHour);
		end;

		saveXMLFile(file);
		delete(file);
	end;
end;

function courseplay_manager:createButton(section, fn, prm, img, x, y, w, h)
	local overlay = Overlay:new(img, Utils.getFilename("img/" .. img, courseplay.path), x, y, w, h);
	local button = { 
		section = section, 
		overlay = overlay, 
		overlays = { overlay }, 
		functionToCall = fn, 
		parameter = prm, 
		x_init = x,
		x = x,
		x2 = (x + w),
		y_init = y,
		y = y,
		y2 = (y + h),
		canBeClicked = fn ~= nil,
		show = true,
		isClicked = false,
		isActive = false,
		isDisabled = false,
		isHovered = false,
		isHidden = false
	};
	--print(string.format("courseplay_manager:createButton(%q, %q, %s, %q, %.3f, %.3f, %.3f, %.3f)", section, fn, prm, img, x, y, w, h));
	table.insert(courseplay_manager.buttons[tostring(section)], button);
	--return #(courseplay_manager.buttons[tostring(section)]);
end;

function courseplay_manager:deleteMap()
	--courses & folders
	g_currentMission.cp_courses = nil
	g_currentMission.cp_folders = nil
	g_currentMission.cp_sorted = nil

	-- debug channels
	for channel,_ in pairs(courseplay.debugChannels) do
		courseplay.debugChannels[channel] = false;
	end;

	--buttons
	for i,vehicle in pairs(g_currentMission.steerables) do
		if vehicle.cp ~= nil then
			if vehicle.cp.globalInfoTextOverlay ~= nil then
				vehicle.cp.globalInfoTextOverlay:delete();
			end;
			if vehicle.cp.buttons ~= nil then
				courseplay.button:deleteButtonOverlays(vehicle);
			end;
		end;
	end;

	--global info text
	for i,button in pairs(self.buttons.globalInfoText) do
		if button.overlays ~= nil then
			for j,overlay in pairs(button.overlays) do
				if overlay.overlayId ~= nil and overlay.delete ~= nil then
					overlay:delete();
				end;
			end;
		end;
		if self.globalInfoTextOverlays[i] then
			local gitOverlay = self.globalInfoTextOverlays[i];
			if gitOverlay.overlayId ~= nil and gitOverlay.delete ~= nil then
				gitOverlay:delete();
			end;
		end;
	end;

	--signs
	for section,signDatas in pairs(courseplay.signs.buffer) do
		for k,signData in pairs(signDatas) do
			courseplay.utils.signs:deleteSign(signData.sign);
		end;
		courseplay.signs.buffer[section] = {};
	end;

	--fields
	courseplay.fields.fieldData = {};
	courseplay.fields.curFieldScanIndex = 0;
	courseplay.fields.allFieldsScanned = false;
	courseplay.fields.ingameDataSetUp = false;
	for i,fruitData in pairs(courseplay.fields.seedUsageCalculator.fruitTypes) do
		if fruitData.overlay then
			fruitData.overlay:delete();
		end;
	end;
	courseplay.fields.seedUsageCalculator = {};
	courseplay.fields.seedUsageCalculator.fieldsWithoutSeedData = {};
end;

function courseplay_manager:draw()
	if g_currentMission.paused then
		return;
	end;

	courseplay.globalInfoText.hasContent = false;
	local git = courseplay.globalInfoText;
	local line = 0;
	local basePosY = courseplay.globalInfoText.backgroundPosY;
	local ingameMap = g_currentMission.ingameMap;
	if not (ingameMap.isVisible and ingameMap.isFullSize) and table.maxn(git.content) > 0 then
		basePosY = ingameMap.isVisible and git.posYAboveMap or git.posY;
		courseplay.globalInfoText.hasContent = true;
		for _,refIndexes in pairs(git.content) do
			if line >= self.globalInfoTextMaxNum then
				break;
			end;

			for refIdx,data in pairs(refIndexes) do
				line = line + 1;

				-- background
				local bg = self.globalInfoTextOverlays[line];
				bg:setColor(unpack(git.levelColors[data.level]));
				local gfxPosY = basePosY + (line - 1) * (git.lineHeight + git.lineMargin);
				bg:setPosition(bg.x, gfxPosY);
				bg:setDimension(data.backgroundWidth, bg.height);
				bg:render();

				-- text
				-- (lineHeight - fontSize) * 0.5
				courseplay:setFontSettings('white', false);
				local textPosY = gfxPosY + (git.lineHeight - git.fontSize); -- should be (lineHeight-fontSize)*0.5, but there seems to be some pixel/sub-pixel rendering error
				renderText(git.textPosX, textPosY, git.fontSize, data.text);

				-- button
				local button = self.buttons.globalInfoText[line];
				if button ~= nil then
					button.overlay:setPosition(button.overlay.x, gfxPosY);
					button.y = gfxPosY;
					button.y2 = gfxPosY + git.buttonHeight;

					local currentColor = button.overlay.curColor;
					local targetColor = currentColor;

					button.canBeClicked = true;
					button.isDisabled = data.vehicle.isBroken or data.vehicle.isControlled;
					button.parameter = data.vehicle;
					if g_currentMission.controlledVehicle and g_currentMission.controlledVehicle == data.vehicle then
						targetColor = 'activeGreen';
						button.canBeClicked = false;
					elseif button.isDisabled then
						targetColor = 'whiteDisabled';
					elseif button.isClicked then
						targetColor = 'activeRed';
					elseif button.isHovered then
						targetColor = 'hover';
					else
						targetColor = 'white';
					end;

					-- set color
					if currentColor ~= targetColor then
						courseplay.button:setButtonColor(button, targetColor)
					end;

					button.overlay:render();
				end;
			end;
		end;
	end;
	self.buttons.globalInfoTextClickArea.y2 = basePosY + (line  * git.lineHeight);

	-- DISPLAY FIELD SCAN MSG
	if courseplay.fields.automaticScan and not courseplay.fields.allFieldsScanned then
		local fsi = self.fieldScanInfo;

		fsi.progressBarBgOverlay:render();
		local pct = courseplay.fields.curFieldScanIndex / g_currentMission.fieldDefinitionBase.numberOfFields;
		local r, g, b = self:getColorFromPct(pct, fsi.percentColors);
		fsi.progressBarOverlay:setColor(r, g, b, 1);
		fsi.progressBarOverlay.width = fsi.progressBarWidth * pct;
  		setOverlayUVs(fsi.progressBarOverlay.overlayId, 0,0, 0,1, pct,0, pct,1);
		fsi.progressBarOverlay:render();

		fsi.bgOverlay:render();

		courseplay:setFontSettings({ 0.8, 0.8, 0.8, 1 }, true, 'left');
		renderText(fsi.lineX, fsi.line1Y - 0.001, courseplay.hud.fontSizes.fieldScanTitle, courseplay:loc('COURSEPLAY_FIELD_SCAN_IN_PROGRESS'));
		courseplay:setFontSettings('shadow', true, 'left');
		renderText(fsi.lineX, fsi.line1Y,         courseplay.hud.fontSizes.fieldScanTitle, courseplay:loc('COURSEPLAY_FIELD_SCAN_IN_PROGRESS'));

		local str2 = courseplay:loc('COURSEPLAY_SCANNING_FIELD_NMB'):format(courseplay.fields.curFieldScanIndex, g_currentMission.fieldDefinitionBase.numberOfFields);
		courseplay:setFontSettings({ 0.8, 0.8, 0.8, 1 }, false, 'left');
		renderText(fsi.lineX, fsi.line2Y - 0.001, courseplay.hud.fontSizes.fieldScanData, str2);
		courseplay:setFontSettings('shadow', false, 'left');
		renderText(fsi.lineX, fsi.line2Y,         courseplay.hud.fontSizes.fieldScanData, str2);

		local rotationStep = math.floor(g_currentMission.time / self.fieldScanInfo.rotationTime);
		if rotationStep > fsi.loadRotStep then
			fsi.loadOverlay:setRotation(rotationStep * fsi.loadRotAdd, fsi.loadOverlay.width/2, fsi.loadOverlay.height/2);
			fsi.loadRotStep = rotationStep;
		end;
		fsi.loadOverlay:render();

		--reset font settings
		courseplay:setFontSettings('white', true, 'left');
	end;
end;

function courseplay_manager:getColorFromPct(pct, colorMap)
	local step = colorMap[2].pct - colorMap[1].pct;

	if pct == 0 then
		return unpack(colorMap[1].color);
	end;

	for i=2, #colorMap do
		local data = colorMap[i];
		if pct == data.pct then
			return unpack(data.color);
		end;

		if pct < data.pct then
			local lower = colorMap[i - 1];
			local upper = colorMap[i];
			local pctAlpha = (pct - lower.pct) / step;
			return Utils.vector3ArrayLerp(lower.color, upper.color, pctAlpha);
		end;
	end;
end;


function courseplay_manager:mouseEvent(posX, posY, isDown, isUp, mouseKey)
	if g_currentMission.paused then return; end;

	local area = courseplay_manager.buttons.globalInfoTextClickArea;
	if area == nil then
		return;
	end;

	--LEFT CLICK
	if (isDown or isUp) and mouseKey == courseplay.inputBindings.mouse.COURSEPLAY_MOUSEACTION.buttonId and courseplay:mouseIsInArea(posX, posY, area.x1, area.x2, area.y1, area.y2) then
		if courseplay.globalInfoText.hasContent then
			for i,button in pairs(self.buttons.globalInfoText) do
				if button.show and courseplay:mouseIsOnButton(posX, posY, button) then
					button.isClicked = isDown;
					if isUp then
						local sourceVehicle = g_currentMission.controlledVehicle or button.parameter;
						courseplay.button:handleMouseClick(sourceVehicle, button);
					end;
					break;
				end;
			end;
		end;

	--RIGHT CLICK
	elseif isDown and mouseKey == courseplay.inputBindings.mouse.COURSEPLAY_MOUSEACTION_SECONDARY.buttonId and g_currentMission.controlledVehicle == nil then
		if courseplay.globalInfoText.hasContent and not self.playerOnFootMouseEnabled then
			self.playerOnFootMouseEnabled = true;
			self.wasPlayerFrozen = g_currentMission.isPlayerFrozen;
			g_currentMission.isPlayerFrozen = true;
		elseif self.playerOnFootMouseEnabled then
			self.playerOnFootMouseEnabled = false;
			if courseplay.globalInfoText.hasContent then --if a button was hovered when deactivating the cursor, deactivate hover state
				for _,button in pairs(self.buttons.globalInfoText) do
					button.isClicked = false;
					button.isHovered = false;
				end;
			end;
			if not self.wasPlayerFrozen then
				g_currentMission.isPlayerFrozen = false;
			end;
		end;
		--print(string.format("right mouse click: playerOnFootMouseEnabled=%s, wasPlayerFrozen=%s, isPlayerFrozen=%s", tostring(self.playerOnFootMouseEnabled), tostring(self.wasPlayerFrozen), tostring(g_currentMission.isPlayerFrozen)));
		InputBinding.setShowMouseCursor(self.playerOnFootMouseEnabled);

	--HOVER
	elseif not isDown and not isUp and courseplay.globalInfoText.hasContent --[[and posX > area.x1 * 0.9 and posX < area.x2 * 1.1 and posY > area.y1 * 0.9 and posY < area.y2 * 1.1]] then
		for _,button in pairs(self.buttons.globalInfoText) do
			button.isClicked = false;
			if button.show and not button.isHidden then
				button.isHovered = false;
				if courseplay:mouseIsOnButton(posX, posY, button) then
					button.isClicked = false;
					button.isHovered = true;
				end;
			end;
		end;
	end;
end;

function courseplay_manager:update(dt)
	if self.firstRun then
		courseplay.moreRealisticInstalled = RealisticUtils ~= nil and RealisticUtils.getFruitInfosV2 ~= nil;
		courseplay.moreRealisticVersion = courseplay:getMoreRealisticVersion();
		courseplay:addCpNilTempFillLevelFunction();

		self.firstRun = false;
	end;

	if g_currentMission.controlledVehicle == nil then
		if self.playerOnFootMouseEnabled then
			g_currentMission:addExtraPrintText(courseplay.inputBindings.mouse.COURSEPLAY_MOUSEACTION_SECONDARY.displayName .. ": " .. courseplay:loc("COURSEPLAY_MOUSEARROW_HIDE"));
		elseif courseplay.globalInfoText.hasContent then
			g_currentMission:addExtraPrintText(courseplay.inputBindings.mouse.COURSEPLAY_MOUSEACTION_SECONDARY.displayName .. ": " .. courseplay:loc("COURSEPLAY_MOUSEARROW_SHOW"));
		end;
	end;

	--SETUP FIELD INGAME DATA
	if not courseplay.fields.ingameDataSetUp then
		courseplay.fields:setUpFieldsIngameData();
	end;

	--SCAN ALL FIELD EDGES
	if courseplay.fields.automaticScan and not courseplay.fields.allFieldsScanned then
		courseplay.fields:setAllFieldEdges();
	end;
end;

function courseplay_manager:updateTick(dt)
end;

function courseplay_manager:keyEvent()
end

function courseplay_manager:load_courses()
	--print("courseplay_manager:load_courses()");
	courseplay:debug('loading courses by courseplay manager', 8);

	local finish_all = false;
	if courseplay.cpXmlFilePath then
		local filePath = courseplay.cpXmlFilePath;

		if fileExists(filePath) then
			local cpFile = loadXMLFile("courseFile", filePath);
			g_currentMission.cp_courses = nil -- make sure it's empty (especially in case of a reload)
			g_currentMission.cp_courses = {}
			local courses_by_id = g_currentMission.cp_courses
			local courses_without_id = {}
			local i = 0
			
			local tempCourse
			repeat

				--current course
				local currentCourse = string.format("XML.courses.course(%d)", i)
				if not hasXMLProperty(cpFile, currentCourse) then
					finish_all = true;
					break;
				end;

				--course name
				local courseName = getXMLString(cpFile, currentCourse .. "#name");
				if courseName == nil then
					courseName = string.format('NO_NAME%d',i)
				end;
				local courseNameClean = courseplay:normalizeUTF8(courseName);

				--course ID
				local id = getXMLInt(cpFile, currentCourse .. "#id")
				if id == nil then
					id = 0;
				end;
				
				--course parent
				local parent = getXMLInt(cpFile, currentCourse .. "#parent")
				if parent == nil then
					parent = 0
				end

				--course waypoints
				tempCourse = {};
				local wpNum = 1;
				local key = currentCourse .. ".waypoint" .. wpNum;
				local finish_wp = not hasXMLProperty(cpFile, key);
				
				while not finish_wp do
					local x, z = Utils.getVectorFromString(getXMLString(cpFile, key .. "#pos"));
					if x ~= nil then
						if z == nil then
							finish_wp = true;
							break;
						end;
						local dangle =   Utils.getVectorFromString(getXMLString(cpFile, key .. "#angle"));
						local wait =     Utils.getVectorFromString(getXMLString(cpFile, key .. "#wait"));
						local speed =    Utils.getVectorFromString(getXMLString(cpFile, key .. "#speed"));
						local rev =      Utils.getVectorFromString(getXMLString(cpFile, key .. "#rev"));
						local crossing = Utils.getVectorFromString(getXMLString(cpFile, key .. "#crossing"));

						--course generation
						local generated =   Utils.getNoNil(getXMLBool(cpFile, key .. "#generated"), false);
						local dir =         getXMLString(cpFile, key .. "#dir");
						local turn =        Utils.getNoNil(getXMLString(cpFile, key .. "#turn"), "false");
						local turnStart =   Utils.getNoNil(getXMLInt(cpFile, key .. "#turnstart"), 0);
						local turnEnd =     Utils.getNoNil(getXMLInt(cpFile, key .. "#turnend"), 0);
						local ridgeMarker = Utils.getNoNil(getXMLInt(cpFile, key .. "#ridgemarker"), 0);

						crossing = crossing == 1 or wpNum == 1;
						wait = wait == 1;
						rev = rev == 1;

						if speed == 0 then
							speed = nil
						end

						--generated not needed, since true or false are loaded from file
						if turn == "false" then
							turn = nil;
						end;
						turnStart = turnStart == 1;
						turnEnd = turnEnd == 1;
						--ridgeMarker not needed, since 0, 1 or 2 is loaded from file

						tempCourse[wpNum] = { 
							cx = x, 
							cz = z, 
							angle = dangle, 
							rev = rev, 
							wait = wait, 
							crossing = crossing, 
							speed = speed,
							generated = generated,
							laneDir = dir,
							turn = turn,
							turnStart = turnStart,
							turnEnd = turnEnd,
							ridgeMarker = ridgeMarker
						};
						
						-- prepare next waypoint
						wpNum = wpNum + 1;
						key = currentCourse .. ".waypoint" .. wpNum;
						finish_wp = not hasXMLProperty(cpFile, key);
					else
						finish_wp = true;
						break;
					end;
				end -- while finish_wp == false;
				
				local course = { id = id, uid = 'c' .. id , type = 'course', name = courseName, nameClean = courseNameClean, waypoints = tempCourse, parent = parent }
				if id ~= 0 then
					courses_by_id[id] = course
				else
					table.insert(courses_without_id, course)
				end
				
				tempCourse = nil;
				i = i + 1;
				
			until finish_all == true;
			
			local j = 0
			local currentFolder, FolderName, id, parent, folder
			finish_all = false
			g_currentMission.cp_folders = nil
			g_currentMission.cp_folders = {}
			local folders_by_id = g_currentMission.cp_folders
			local folders_without_id = {}
			repeat
				-- current folder
				currentFolder = string.format("XML.folders.folder(%d)", j)
				if not hasXMLProperty(cpFile, currentFolder) then
					finish_all = true;
					break;
				end;
				
				-- folder name
				FolderName = getXMLString(cpFile, currentFolder .. "#name")
				if FolderName == nil then
					FolderName = string.format('NO_NAME%d',j)
				end
				local folderNameClean = courseplay:normalizeUTF8(FolderName);
				
				-- folder id
				id = getXMLInt(cpFile, currentFolder .. "#id")
				if id == nil then
					id = 0
				end
				
				-- folder parent
				parent = getXMLInt(cpFile, currentFolder .. "#parent")
				if parent == nil then
					parent = 0
				end
				
				-- "save" current folder
				folder = { id = id, uid = 'f' .. id ,type = 'folder', name = FolderName, nameClean = folderNameClean, parent = parent }
				if id ~= 0 then
					folders_by_id[id] = folder
				else
					table.insert(folders_without_id, folder)
				end
				j = j + 1
			until finish_all == true
			
			local save = false
			if #courses_without_id > 0 then
				-- give a new ID and save
				local maxID = courseplay.courses.getMaxCourseID()
				for i = 1, #courses_without_id do
					maxID = maxID + 1
					courses_without_id[i].id = maxID
					courses_without_id[i].uid = 'c' .. maxID
					courses_by_id[maxID] = courses_without_id[i]
				end
				save = true
			end
			if #folders_without_id > 0 then
				-- give a new ID and save
				local maxID = courseplay.courses.getMaxFolderID()
				for i = #folders_without_id, 1, -1 do
					maxID = maxID + 1
					folders_without_id[i].id = maxID
					folders_without_id[i].uid = 'f' .. maxID
					folders_by_id[maxID] = table.remove(folders_without_id)
				end
				save = true
			end		
			if save then
				-- this will overwrite the courseplay file and therefore delete the courses without ids and add them again with ids as they are now stored in g_currentMission with an id
				courseplay.courses.save_all()
			end
			
			g_currentMission.cp_sorted = courseplay.courses.sort(courses_by_id, folders_by_id, 0, 0)
						
			delete(cpFile);
		else
			-- print(('\t"courseplay.xml" missing at %q'):format(tostring(courseplay.cpXmlFilePath);
		end; --END if fileExists
		
		courseplay:debug(tableShow(g_currentMission.cp_sorted.item, "cp_sorted.item", 8), 8);

		return g_currentMission.cp_courses;
	else
		print(('Error: [Courseplay] current savegame could not be found at %q.'):format(courseplay.cpXmlFilePath));
	end; --END if savegame ~= nil

	return nil;
end


--remove courseplayers from combine / combine from tractor before it is reset and/or sold
function courseplay_manager:severCombineTractorConnection(vehicle, callDelete)
	if vehicle.cp then
		-- VEHICLE IS COMBINE
		if vehicle.cp.isCombine or vehicle.cp.isChopper or vehicle.cp.isHarvesterSteerable or vehicle.cp.isSugarBeetLoader or courseplay:isSpecialChopper(vehicle) then
			courseplay:debug(('BaseMission:removeVehicle() -> severCombineTractorConnection(%q, %s) [VEHICLE IS COMBINE]'):format(nameNum(vehicle), tostring(callDelete)), 4);
			local combine = vehicle;
			-- remove this combine as savedCombine from all tractors
			for i,tractor in pairs(g_currentMission.steerables) do
				if tractor.cp and tractor.cp.savedCombine and tractor.cp.savedCombine == combine then
					courseplay:debug(('\ttractor %q: savedCombine=%q --> removeSavedCombineFromTractor()'):format(nameNum(tractor), nameNum(combine)), 4);
					courseplay:removeSavedCombineFromTractor(tractor);
				end;
			end;

			-- unregister all tractors from this combine (activeCombine)
			if combine.courseplayers ~= nil then
				courseplay:debug(('\t.courseplayers ~= nil (%d courseplayers)'):format(#combine.courseplayers), 4);
				if #combine.courseplayers > 0 then
					for i,tractor in pairs(combine.courseplayers) do
						courseplay:debug(('\t\t%q: removeActiveCombineFromTractor(), removeSavedCombineFromTractor()'):format(nameNum(tractor)), 4);
						courseplay:removeActiveCombineFromTractor(tractor);
						courseplay:removeSavedCombineFromTractor(tractor); --TODO (Jakob): unnecessary, as done above in steerables table already?
						tractor.cp.reachableCombines = nil;
					end;
					courseplay:debug(('\t-> now has %d courseplayers'):format(#combine.courseplayers), 4);
				end;
			end;

		-- VEHICLE IS TRACTOR
		elseif vehicle.cp.activeCombine ~= nil or vehicle.cp.lastActiveCombine ~= nil or vehicle.cp.savedCombine ~= nil then
			courseplay:debug(('BaseMission:removeVehicle() -> severCombineTractorConnection(%q, %s) [VEHICLE IS TRACTOR]'):format(nameNum(vehicle), tostring(callDelete)), 4);
			courseplay:debug(('\tactiveCombine=%q, lastActiveCombine=%q, savedCombine=%q -> removeActiveCombineFromTractor(), removeSavedCombineFromTractor()'):format(nameNum(vehicle.cp.activeCombine), nameNum(vehicle.cp.lastActiveCombine), nameNum(vehicle.cp.savedCombine)), 4);
			courseplay:removeActiveCombineFromTractor(vehicle);
			courseplay:removeSavedCombineFromTractor(vehicle);
			courseplay:debug(('\t-> activeCombine=%q, lastActiveCombine=%q, savedCombine=%q'):format(nameNum(vehicle.cp.activeCombine), nameNum(vehicle.cp.lastActiveCombine), nameNum(vehicle.cp.savedCombine)), 4);
		end;
	end;
end;
BaseMission.removeVehicle = Utils.prependedFunction(BaseMission.removeVehicle, courseplay_manager.severCombineTractorConnection);

function courseplay_manager:devAddMoney()
	if g_server ~= nil then
		g_currentMission:addSharedMoney(5000000, 'other');
		return ('Added %s to your bank account'):format(g_i18n:formatMoney(5000000));
	end;
end;
function courseplay_manager:devAddFillLevels()
	if g_server ~= nil then
		for fillType=1,Fillable.NUM_FILLTYPES do
			g_currentMission:setSiloAmount(fillType, g_currentMission:getSiloAmount(fillType) + 500000);
		end;
		return 'All silo fill levels increased by 500\'000.';
	end;
end;

local nightStart, dayStart = 19 * 3600000, 7.5 * 3600000; -- from 7pm until 7:30am
function courseplay_manager:minuteChanged()
	-- WEATHER
	local env = g_currentMission.environment;
	courseplay.lightsNeeded = env.needsLights or (env.dayTime >= nightStart or env.dayTime <= dayStart) or env.currentRain ~= nil or env.curRain ~= nil or (env.lastRainScale > 0.1 and env.timeSinceLastRain < 30);

	-- WAGES
	if courseplay.wagesActive and g_server ~= nil then
		local totalWages = 0;
		for vehicleNum, vehicle in ipairs(courseplay.totalCoursePlayers) do
			if vehicle.drive and not vehicle.isHired then
				totalWages = totalWages + courseplay.wagePerMin;
			end;
		end;
		if totalWages > 0 then
			g_currentMission:addSharedMoney(-totalWages * courseplay.wageDifficultyMultiplier, 'wagePayment');
		end;
	end;
end;

addModEventListener(courseplay_manager);



stream_debug_counter = 0


--
-- based on PlayerJoinFix
--
-- SFM-Modding
-- @author:  Manuel Leithner
-- @date:    01/08/11
-- @version: v1.1
-- @history: v1.0 - initial implementation
--           v1.1 - adaption to courseplay
--

local modName = g_currentModName;
local Server_sendObjects_old = Server.sendObjects;

function Server:sendObjects(connection, x, y, z, viewDistanceCoeff)
	connection:sendEvent(CourseplayJoinFixEvent:new());

	Server_sendObjects_old(self, connection, x, y, z, viewDistanceCoeff);
end


CourseplayJoinFixEvent = {};
CourseplayJoinFixEvent_mt = Class(CourseplayJoinFixEvent, Event);

InitEventClass(CourseplayJoinFixEvent, "CourseplayJoinFixEvent");

function CourseplayJoinFixEvent:emptyNew()
	local self = Event:new(CourseplayJoinFixEvent_mt);
	self.className = modName .. ".CourseplayJoinFixEvent";
	return self;
end

function CourseplayJoinFixEvent:new()
	local self = CourseplayJoinFixEvent:emptyNew()
	return self;
end

function CourseplayJoinFixEvent:writeStream(streamId, connection)


	if not connection:getIsServer() then
		--courseplay:debug("manager transfering courses", 8);
		--transfer courses
		local course_count = 0
		for _,_ in pairs(g_currentMission.cp_courses) do
			course_count = course_count + 1
		end
		streamDebugWriteInt32(streamId, course_count)
		print(string.format("\t### CourseplayMultiplayer: writing %d courses ", course_count ))
		for id, course in pairs(g_currentMission.cp_courses) do
			streamDebugWriteString(streamId, course.name)
			streamDebugWriteString(streamId, course.uid)
			streamDebugWriteString(streamId, course.type)
			streamDebugWriteInt32(streamId, course.id)
			streamDebugWriteInt32(streamId, course.parent)
			streamDebugWriteInt32(streamId, #(course.waypoints))
			for w = 1, #(course.waypoints) do
				streamDebugWriteFloat32(streamId, course.waypoints[w].cx)
				streamDebugWriteFloat32(streamId, course.waypoints[w].cz)
				streamDebugWriteFloat32(streamId, course.waypoints[w].angle)
				streamDebugWriteBool(streamId, course.waypoints[w].wait)
				streamDebugWriteBool(streamId, course.waypoints[w].rev)
				streamDebugWriteBool(streamId, course.waypoints[w].crossing)
				streamDebugWriteInt32(streamId, course.waypoints[w].speed)

				streamDebugWriteBool(streamId, course.waypoints[w].generated)
				streamDebugWriteString(streamId, (course.waypoints[w].laneDir or ""))
				streamDebugWriteString(streamId, course.waypoints[w].turn)
				streamDebugWriteBool(streamId, course.waypoints[w].turnStart)
				streamDebugWriteBool(streamId, course.waypoints[w].turnEnd)
				streamDebugWriteInt32(streamId, course.waypoints[w].ridgeMarker)
			end
		end
				
		local folderCount = 0
		for _,_ in pairs(g_currentMission.cp_folders) do
			folderCount = folderCount + 1
		end
		streamDebugWriteInt32(streamId, folderCount)
		print(string.format("\t### CourseplayMultiplayer: writing %d folders ", folderCount ))
		for id, folder in pairs(g_currentMission.cp_folders) do
			streamDebugWriteString(streamId, folder.name)
			streamDebugWriteString(streamId, folder.uid)
			streamDebugWriteString(streamId, folder.type)
			streamDebugWriteInt32(streamId, folder.id)
			streamDebugWriteInt32(streamId, folder.parent)
		end
				
		local fieldsCount = 0
		for _, field in pairs(courseplay.fields.fieldData) do
			if field.isCustom then
				fieldsCount = fieldsCount+1
			end
		end
		streamDebugWriteInt32(streamId, fieldsCount)
		print(string.format("\t### CourseplayMultiplayer: writing %d custom fields ", fieldsCount))
		for id, course in pairs(courseplay.fields.fieldData) do
			if course.isCustom then
				streamDebugWriteString(streamId, course.name)
				streamDebugWriteInt32(streamId, course.numPoints)
				streamDebugWriteBool(streamId, course.isCustom)
				streamDebugWriteInt32(streamId, course.fieldNum)
				streamDebugWriteInt32(streamId, #(course.points))
				for p = 1, #(course.points) do
					streamDebugWriteFloat32(streamId, course.points[p].cx)
					streamDebugWriteFloat32(streamId, course.points[p].cy)
					streamDebugWriteFloat32(streamId, course.points[p].cz)
				end
			end
		end
	end;
end

function CourseplayJoinFixEvent:readStream(streamId, connection)
	if connection:getIsServer() then
		local course_count = streamDebugReadInt32(streamId)
		print(string.format("\t### CourseplayMultiplayer: reading %d couses ", course_count ))
		g_currentMission.cp_courses = {}
		for i = 1, course_count do
			--courseplay:debug("got course", 8);
			local course_name = streamDebugReadString(streamId)
			local courseUid = streamDebugReadString(streamId)
			local courseType = streamDebugReadString(streamId)
			local course_id = streamDebugReadInt32(streamId)
			local courseParent = streamDebugReadInt32(streamId)
			local wp_count = streamDebugReadInt32(streamId)
			local waypoints = {}
			for w = 1, wp_count do
				--courseplay:debug("got waypoint", 8);
				local cx = streamDebugReadFloat32(streamId)
				local cz = streamDebugReadFloat32(streamId)
				local angle = streamDebugReadFloat32(streamId)
				local wait = streamDebugReadBool(streamId)
				local rev = streamDebugReadBool(streamId)
				local crossing = streamDebugReadBool(streamId)
				local speeed = streamDebugReadInt32(streamId)

				local generated = streamDebugReadBool(streamId)
				local dir = streamDebugReadString(streamId)
				local turn = streamDebugReadString(streamId)
				local turnStart = streamDebugReadBool(streamId)
				local turnEnd = streamDebugReadBool(streamId)
				local ridgeMarker = streamDebugReadInt32(streamId)
				
				local wp = {
					cx = cx, 
					cz = cz, 
					angle = angle, 
					wait = wait, 
					rev = rev, 
					crossing = crossing, 
					speed = speed,
					generated = generated,
					laneDir = dir,
					turn = turn,
					turnStart = turnStart,
					turnEnd = turnEnd,
					ridgeMarker = ridgeMarker 
				};
				table.insert(waypoints, wp)
			end
			local course = { id = course_id, uid = courseUid, type = courseType, name = course_name, nameClean = courseplay:normalizeUTF8(course_name), waypoints = waypoints, parent = courseParent }
			g_currentMission.cp_courses[course_id] = course
			g_currentMission.cp_sorted = courseplay.courses.sort()
		end
		
		local folderCount = streamDebugReadInt32(streamId)
		print(string.format("\t### CourseplayMultiplayer: reading %d folders ", folderCount ))
		g_currentMission.cp_folders = {}
		for i = 1, folderCount do
			local folderName = streamDebugReadString(streamId)
			local folderUid = streamDebugReadString(streamId)
			local folderType = streamDebugReadString(streamId)
			local folderId = streamDebugReadInt32(streamId)
			local folderParent = streamDebugReadInt32(streamId)
			local folder = { id = folderId, uid = folderUid, type = folderType, name = folderName, nameClean = courseplay:normalizeUTF8(folderName), parent = folderParent }
			g_currentMission.cp_folders[folderId] = folder
			g_currentMission.cp_sorted = courseplay.courses.sort(g_currentMission.cp_courses, g_currentMission.cp_folders, 0, 0)
		end
		
		local fieldsCount = streamDebugReadInt32(streamId)		
		print(string.format("\t### CourseplayMultiplayer: reading %d custom fields ", fieldsCount))
		courseplay.fields.fieldData = {}
		for i = 1, fieldsCount do
			local name = streamDebugReadString(streamId)
			local numPoints = streamDebugReadInt32(streamId)
			local isCustom = streamDebugReadBool(streamId)
			local fieldNum = streamDebugReadInt32(streamId)
			local ammountPoints = streamDebugReadInt32(streamId)
			local waypoints = {}
			for w = 1, ammountPoints do 
				local cx = streamDebugReadFloat32(streamId)
				local cy = streamDebugReadFloat32(streamId)
				local cz = streamDebugReadFloat32(streamId)
				local wp = { cx = cx, cy = cy, cz = cz}
				table.insert(waypoints, wp)
			end
			local field = { name = name, numPoints = numPoints, isCustom = isCustom, fieldNum = fieldNum, points = waypoints}
			courseplay.fields.fieldData[fieldNum] = field
		end
		print("\t### CourseplayMultiplayer: courses/folders reading end")
	end;
end

function CourseplayJoinFixEvent:run(connection)
	--courseplay:debug("CourseplayJoinFixEvent Run function should never be called", 8);
end;